__doc__='''Shopper's Basket implementation'''

__version__='$Revision: 1.26 $'[11:-2]

BasketError = "Basket Error."

from Globals import HTMLFile, MessageDialog, Persistent
import OFS.SimpleItem, Acquisition
import traceback

from Persistence import PersistentMapping
    
import string
import time

ORDER_PENDING, ORDER_SUBMITTED, ORDER_VALIDATED, ORDER_SHIPPED, ORDER_COMPLETE = range(5)
state_strings = ["Pending", "Submitted", "Validated", "Shipped", "Complete"]

class State(Persistent):
    """ State is a timestamped condition..."""

    state = None
    time = None

    def __init__(self, theState = None):
        self.time = time.time()
        if theState is None:
            self.state = ORDER_PENDING
        else:
            self.state = theState

    def __str__(self):
        return state_strings[self.state]
    
class Order(Persistent):
    """ Order keeps track of a single order and it's current state. """
    
    items = None
    states = []
    transID = None

    def __init__(self, transID):
        self.transID = transID
        self.items = PersistentMapping()
        self.states = [State()]

    def UpdateTransID(self, transID):
        self.transID = transID

    def __setitem__(self, key, val):
        self.items[key] = val

    def __getitem__(self, key):
        return self.items[key]

    def __delitem__(self, key):
        del self.items[key]

    def has_key(self, key):
        return self.items.has_key(key)

    def keys(self):
        return self.items.keys()

    def values(self):
        return self.items.values()

    def items(self):
        return self.items.items()

class BasketItem(Persistent):

    sku = ''
    quantity = 0
    offer = None
    counterOffer = None
    counterQuantity = None
    optionDict = None
    isBasketItem = 1

    def __init__(self, aMarketItem, quantity, optionDict=None):
        self.sku = aMarketItem.sku
        self.quantity = quantity
        self.optionDict = optionDict

    def optionDict2Mapping(self):
        """ convert self.optionDict to a mapping object for display..."""
        first = string.find(self.optionDict,'(')
        last = string.rfind(self.optionDict,')')
        innerString = self.optionDict[first+1:last]  # should be '(a,b),(c,d),....'
        pairs = string.split(innerString,',')
        newMapping = {}
        for item in range(len(pairs)/2):
            [key,val] = pairs[2*item:2*item+2]
            key = string.strip(key)
            val = string.strip(val)
            if key[0] == '(':
                key = key[1:]
            if key[0] == "'":
                key = key[1:]
            if key[-1] == "'":
                key = key[:-1]
            if val[-1] == ')':
                val = val[:-1]
            if val[0] == "'":
                val = val[1:]
            if val[-1] == "'":
                val = val[:-1]

            newMapping[key] = val

        return newMapping
        
class Basket( Persistent, Acquisition.Implicit ):

    shopper = None
    
    def __init__(self, shopper = None ):
        self.shopper = shopper
        self.items = None
        self.clearProcessedItems()

    def UpdateShopper(self, shopper):
        """ we have a new shopper..."""
        self.shopper = shopper
        self.items.UpdateTransID( shopper.transID )

    def ClearTransaction(self):
        """ Clear out old basket/transaction """
        self.items = Order(self.shopper.transID)

    def getItem(self, anItem, optionDict = None):
        """ get an item from the basket """

        theKey = anItem.id

        if optionDict is not None:
            if type(optionDict) == type(''):
                theKey = optionDict
            else:
                theList = []
                keys = optionDict.keys()
                keys.sort()
                for key in keys:
                    theList.append((key,optionDict[key]))

                optionDict = tuple(theList)
                theKey = theKey + `optionDict`
                
        return self.items[theKey]
        
    def addItem(self, anItem, quantity=1, updateBasket=None, optionDict = None, offer=None):
        """ add to/set the quantity for an item in the basket """

        theKey = anItem.id

        if optionDict is not None:
            if type(optionDict) == type(''):
                theKey = optionDict
            else:
                theList = []
                keys = optionDict.keys()
                keys.sort()
                for key in keys:
                    theList.append((key,optionDict[key]))

                optionDict = tuple(theList)
                theKey = theKey + `optionDict`
                
        if type(quantity) == type(''):
            quantity = int(string.atof(quantity))

        if type(offer) == type(''):
            offer = string.strip(offer)
            offer = string.replace(offer,',','')
            offer = string.replace(offer,'$','')
            try:
                offer = string.atof(offer)
            except:
                offer = None

        if (updateBasket is not None) and (quantity <= 0):
            # we are getting rid of this item.....
            #
            self.deleteItem( theKey = theKey )

        else:

            if self.items is None:
                self.ClearTransaction()
                
            i = self.items

            if i.has_key(theKey):
                if updateBasket is not None:
                    i[theKey].quantity = quantity
                else:
                    i[theKey].quantity = i[theKey].quantity + quantity
            else:
                i[theKey] = BasketItem(anItem, quantity, theKey)

            i[theKey].offer = offer
                
            self.items = i

    def calcTotal(self):
        """ calculate the total of the items in the basket...."""

        sum = 0.0

        for k in self.items.keys():
            theBasketItem = self.items[k]
            quantity = theBasketItem.quantity
            theItem = self.shopper.getMarketItem(theBasketItem.sku)
            if theItem:
                price = theItem.calcPrice(theBasketItem.optionDict, theBasketItem)
            sum = sum + price*quantity

        return sum

    def deleteItem(self, theKey = None):
        """ delete an item from the basket."""

        i = self.items

        if theKey is not None:
            if i.has_key(theKey):
                del i[theKey]
                self.items = i


    def processItem(self, item):
        """

        Process a single basket item rather than an 'order...'

        """
        #
        pi = self.processedItems
        theKey = self.shopper.id + '-' + `time.time()`
        newOrder = Order(theKey)
        newOrder[theKey] = item
        newOrder.states.append(State(ORDER_COMPLETE))
        pi[theKey] = newOrder
        self.processItems = pi

    def clearProcessedItems(self):
        """ clear out the processed items...."""

        self.processedItems = PersistentMapping()
        
    def processTransaction(self, transID=None):
        """
        Set the 'processed' state of the basket. Copy any current items
        to 'processedItems'. This is for shopping areas that process
        one 'order' at a time. Auction sites process one item at
        a time and should use 'processItem'.
        
        """

        pi = self.processedItems
        pi[transID] = self.items           # save the old transaction..
        self.processedItems = pi

        self.shopper.createTransID(makeNew = 1)   # This is a new transaction...

    def purgeBasket(self):
        """ We need to eliminate cruft... (items that are no longer valid)"""

        purgeList = []
        
        for k in self.items.keys():
            theBasketItem = self.items[k]
            marketItem = self.getMarketItem(theBasketItem.sku)
            if marketItem:
                if not hasattr(self.shopperFolder, marketItem.shopperID):
                    self.deleteItem(k)
                    purgeList.append(k)

            else:
                self.deleteItem(k)
                purgeList.append(k)

        return purgeList
